;; Generated package description from substitute.el  -*- no-byte-compile: t -*-
(define-package "substitute" "0.4.0.0.20251129.84101" "Efficiently replace targets in the buffer or context" '((emacs "27.1")) :commit "63818140b432212ea2d5314a27469c53535b61f8" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/substitute")
